from __future__ import annotations

import ast
from collections.abc import Callable, Generator
from contextlib import contextmanager, nullcontext
from dataclasses import dataclass, field
from functools import reduce, wraps
from itertools import chain
from typing import Any

import pytest

from pytest_smoke.plugin import DEFAULT_N, SmokeScope
from pytest_smoke.utils import scale_down

TEST_NAME_BASE = "test_something"


@dataclass
class TestFileSpec:
    """Specification of a test file"""

    __test__ = False

    test_specs: list[TestFuncSpec | TestClassSpec] = field(default_factory=list)
    test_dir: str | None = None


@dataclass
class TestClassSpec:
    """Specification of a test class"""

    __test__ = False

    name: str
    test_func_specs: list[TestFuncSpec] = field(default_factory=list)
    num_params: int = 0
    param_marker: Callable[[int], str | None] | None = None
    nested_test_class_specs: list[TestClassSpec] = field(default_factory=list)
    parent_test_class_spec: TestClassSpec | None = field(init=False, default=None)
    param_arg_name: str = field(init=False, default="")

    def __post_init__(self) -> None:
        if self.num_params:
            self.param_arg_name = f"c{COUNTER.get_next()}"
        for test_func_spec in self.test_func_specs:
            test_func_spec.test_class_spec = self
        for nested_test_class_spec in self.nested_test_class_specs:
            nested_test_class_spec.parent_test_class_spec = self

    @property
    def param_arg_names(self) -> list[str]:
        def get_arg_names(class_spec: TestClassSpec, arg_names: list[str] | None = None) -> list[str]:
            if arg_names is None:
                arg_names = []
            if arg_name := class_spec.param_arg_name:
                arg_names.append(arg_name)
            if class_spec.parent_test_class_spec:
                get_arg_names(class_spec.parent_test_class_spec, arg_names=arg_names)
            return arg_names

        return get_arg_names(self)

    @property
    def has_parametrized_test(self) -> bool:
        def is_test_class_parametrized(test_class_spec: TestClassSpec) -> bool:
            return test_class_spec.num_params > 0 or (
                test_class_spec.parent_test_class_spec is not None
                and is_test_class_parametrized(test_class_spec.parent_test_class_spec)
            )

        if is_test_class_parametrized(self):
            return True

        for test_spec in self.test_func_specs:
            if test_spec.num_params:
                return True
        return False


@dataclass
class TestFuncSpec:
    """Specification of a test function"""

    __test__ = False
    param_arg_name = "p"

    num_params: int = 0
    test_class_spec: TestClassSpec | None = field(init=False, default=None)
    param_marker: Callable[[int], str | None] | None = None
    func_body: str | None = None

    @property
    def has_parametrized_test(self) -> bool:
        if self.num_params > 0 or (self.test_class_spec and self.test_class_spec.has_parametrized_test):
            return True
        return False


class Counter:
    def __init__(self) -> None:
        self._count = 0

    def get_next(self) -> int:
        self._count += 1
        return self._count


COUNTER = Counter()


def generate_test_code(test_spec: TestFileSpec | TestFuncSpec) -> str:
    """Generate test code from the test file spec

    :param test_spec: Test file spec or Test func spec
    """
    code = "import pytest\n"

    def add_parametrize_marker(
        *param_names: str,
        num_params: int = 1,
        param_marker: Callable[[int], str | None] | None = None,
        indent: int = 0,
    ) -> None:
        nonlocal code
        params = ", ".join(param_names)
        tabs = "\t" * indent
        if param_marker:
            param_values = ", ".join(
                [
                    (f"pytest.param({p}, marks=pytest.mark.{param_marker(p)})" if param_marker(p) else str(p))
                    for p in range(num_params)
                ]
            )
            code += f"{tabs}@pytest.mark.parametrize('{params}', [{param_values}])\n"
        else:
            code += f"{tabs}@pytest.mark.parametrize('{params}', range({num_params}))\n"

    def add_class(test_class_spec: TestClassSpec, indent: int = 0) -> None:
        nonlocal code
        if test_class_spec.num_params:
            add_parametrize_marker(
                test_class_spec.param_arg_name,
                num_params=test_class_spec.num_params,
                param_marker=test_class_spec.param_marker,
                indent=indent,
            )

        tabs = "\t" * indent
        code += f"{tabs}class {test_class_spec.name}:\n"
        for j, func_spec in enumerate(test_class_spec.test_func_specs, start=1):
            add_function(f"{TEST_NAME_BASE}{j}", func_spec, indent=indent + 1)

        for nested_test_class in test_class_spec.nested_test_class_specs:
            add_class(nested_test_class, indent=indent + 1)

    def add_function(name: str, test_func_spec: TestFuncSpec, indent: int = 0) -> None:
        nonlocal code
        params: list[str] = []
        test_class_spec = test_func_spec.test_class_spec
        if test_class_spec and test_class_spec.param_arg_names:
            params.extend(test_class_spec.param_arg_names)
        if test_func_spec.num_params:
            params.append(test_func_spec.param_arg_name)
            add_parametrize_marker(
                test_func_spec.param_arg_name,
                num_params=test_func_spec.num_params,
                param_marker=test_func_spec.param_marker,
                indent=indent,
            )
        if test_class_spec:
            params = ["self", *params]
        tabs = "\t" * indent
        func_args = ", ".join(params)
        func_body = "\n".join(f"{tabs}\t{x}" for x in (test_func_spec.func_body or "pass\n").split("\n"))
        code += f"{tabs}def {name}({func_args}):\n{func_body}\n"

    if isinstance(test_spec, TestFileSpec):
        for i, spec in enumerate(test_spec.test_specs, start=1):
            if isinstance(spec, TestClassSpec):
                add_class(spec)
            else:
                add_function(f"{TEST_NAME_BASE}{i}", spec)
    else:
        add_function(TEST_NAME_BASE, test_spec)

    try:
        ast.parse(code)
    except Exception as e:
        raise RuntimeError(
            f"Failed to parse the generated code:\nError: {type(e).__name__} {e}\nGenerated code:\n{code}"
        )
    return code


def get_num_tests(*test_specs: TestFuncSpec | TestClassSpec | TestFileSpec, include_nested_classes: bool = True) -> int:
    """Calculate the total number of tests from the given test specs

    :param test_specs: One or more TestFuncSpec, TestClassSpec, or TestFileSpec objects to calculate the total number of
                       tests against
    :param include_nested_classes: Count tests defined inside nested classes, if any
    """

    def _get_num_tests(test_func_spec: TestFuncSpec) -> int:
        def collect_num_params_on_test_class(current_class_spec: TestClassSpec) -> None:
            num_params.append(current_class_spec.num_params or 1)
            if current_class_spec.parent_test_class_spec:
                collect_num_params_on_test_class(current_class_spec.parent_test_class_spec)

        num_params = [test_func_spec.num_params or 1]
        if test_func_spec.test_class_spec:
            collect_num_params_on_test_class(test_func_spec.test_class_spec)
        return reduce(lambda x, y: x * y, num_params)

    num_tests = 0
    for spec in test_specs:
        if isinstance(spec, TestFileSpec):
            num_tests += get_num_tests(*spec.test_specs, include_nested_classes=include_nested_classes)
        elif isinstance(spec, TestClassSpec):
            num_tests += get_num_tests(*spec.test_func_specs, include_nested_classes=include_nested_classes)
            if include_nested_classes and spec.nested_test_class_specs:
                num_tests += get_num_tests(*spec.nested_test_class_specs, include_nested_classes=include_nested_classes)
        else:
            num_tests += _get_num_tests(spec)
    return num_tests


def get_num_tests_to_be_selected(test_file_specs: list[TestFileSpec], n: str | None, scope: SmokeScope | None) -> int:
    """Calculate the number of expected tests to be selected from the test file specs for the given conditions

    :param test_file_specs: Test file specs
    :param n: N value given for the smoke option
    :param scope: Smoke scope value
    """
    scopes_handle_class_specs = [None, SmokeScope.CLASS, SmokeScope.AUTO]

    def get_num_expected_tests(
        *test_specs: TestFuncSpec | TestClassSpec | TestFileSpec, effective_scope: SmokeScope | None = scope
    ) -> int:
        if effective_scope == SmokeScope.AUTO:
            raise ValueError("Must give one of the non-Auto scopes")

        n_ = n or str(DEFAULT_N)
        include_nested_classes = effective_scope not in scopes_handle_class_specs
        if n_.endswith("%"):
            scale = float(n_[:-1])
            return int(scale_down(get_num_tests(*test_specs, include_nested_classes=include_nested_classes), scale))
        else:
            return min([int(n_), get_num_tests(*test_specs, include_nested_classes=include_nested_classes)])

    if scope == SmokeScope.ALL:
        num_expected_tests = get_num_expected_tests(*test_file_specs)
    elif scope == SmokeScope.DIRECTORY:
        test_specs_per_dir: dict[str | None, list[TestFuncSpec | TestClassSpec]] = {}
        for test_file_spec in test_file_specs:
            test_specs_per_dir.setdefault(test_file_spec.test_dir, []).extend(test_file_spec.test_specs)
        num_expected_tests = sum(get_num_expected_tests(*test_specs) for test_specs in test_specs_per_dir.values())
    elif scope == SmokeScope.FILE:
        num_expected_tests = sum(get_num_expected_tests(*x.test_specs) for x in test_file_specs)
    elif scope == SmokeScope.CLASS:
        test_class_specs = get_test_class_specs(test_file_specs)
        num_expected_tests = sum(get_num_expected_tests(x) for x in test_class_specs)
    elif scope == SmokeScope.FUNCTION:
        num_expected_tests = sum(get_num_expected_tests(x) for x in get_test_func_specs(test_file_specs))
    else:
        # default scope = auto
        num_expected_tests = 0
        for test_file_spec in test_file_specs:
            test_func_specs, test_class_specs = _get_test_specs_per_type([test_file_spec])
            if test_func_specs:
                if any(x.has_parametrized_test for x in test_func_specs):
                    num_expected_tests += sum(
                        get_num_expected_tests(x, effective_scope=SmokeScope.FUNCTION) for x in test_func_specs
                    )
                else:
                    num_expected_tests += get_num_expected_tests(*test_func_specs, effective_scope=SmokeScope.FILE)
            if test_class_specs:
                for test_class_spec in test_class_specs:
                    if test_class_spec.has_parametrized_test:
                        num_expected_tests += sum(
                            get_num_expected_tests(x, effective_scope=SmokeScope.FUNCTION)
                            for x in test_class_spec.test_func_specs
                        )
                    else:
                        num_expected_tests += get_num_expected_tests(test_class_spec, effective_scope=SmokeScope.CLASS)
    return num_expected_tests


def get_test_class_specs(test_file_specs: list[TestFileSpec]) -> list[TestClassSpec]:
    """Returns all TestClassSpec objects from test file specs

    :param test_file_specs: Test file specs
    """
    _, test_class_specs = _get_test_specs_per_type(test_file_specs)
    return test_class_specs


def get_test_func_specs(test_file_specs: list[TestFileSpec], exclude_class: bool = False) -> list[TestFuncSpec]:
    """Returns all TestFuncSpec objects from test file specs

    :param test_file_specs: Test file specs
    :param exclude_class: Exclude test func specs defined under a test class spec
    """
    test_func_specs, test_class_specs = _get_test_specs_per_type(test_file_specs)
    if exclude_class:
        return test_func_specs
    else:
        test_class_func_specs = list(chain(*[x.test_func_specs for x in test_class_specs]))
        return [*test_func_specs, *test_class_func_specs]


def _get_test_specs_per_type(test_file_specs: list[TestFileSpec]) -> tuple[list[TestFuncSpec], list[TestClassSpec]]:
    test_specs = list(chain(*[x.test_specs for x in test_file_specs]))
    test_class_specs = [x for x in test_specs if isinstance(x, TestClassSpec)]
    nested_test_class_specs = list(chain(*[x.nested_test_class_specs for x in test_class_specs]))
    func_specs = [x for x in test_specs if isinstance(x, TestFuncSpec)]
    return (func_specs, [*test_class_specs, *nested_test_class_specs])


@contextmanager
def mock_column_width(width: int) -> Generator[None, Any, None]:
    """Temporarily mock the column width

    :param width: Column width
    """
    mp = pytest.MonkeyPatch()
    mp.setenv("COLUMNS", str(width))
    try:
        yield
    finally:
        mp.undo()


def patch_runpytest(f: Callable[..., Any]) -> Callable[..., Any]:
    """A decorator for patching pytester.runpytest() to temporarily mock the column width during a test when the -v
    option is given to ensure the stdout is captured in a standard terminal size.
    This prevents the flaky test results caused by a teser's actual terminal size
    """

    @wraps(f)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        with mock_column_width(150) if "-v" in args else nullcontext():
            return f(*args, **kwargs)

    return wrapper



from __future__ import annotations

import os
import random
from collections.abc import Callable, Generator
from contextlib import contextmanager
from decimal import ROUND_HALF_UP, Decimal
from functools import cache, update_wrapper
from types import ModuleType
from typing import TYPE_CHECKING, Any, ClassVar, cast
from uuid import UUID

import pytest
from pytest import Class, Function

from pytest_smoke import smoke
from pytest_smoke.types import SmokeEnvVar, SmokeIniOption, SmokeOption, SmokeScope, SmokeSelectMode

if smoke.is_xdist_installed:
    from xdist import is_xdist_controller, is_xdist_worker

if TYPE_CHECKING:
    from _pytest.nodes import Node
    from pytest import Config, Item, Session


class Cache:
    """Custom functools.cache decorator that provides an easy way to clear cache while allowing unlimited cache size"""

    _cached_wrappers: ClassVar[set[Callable[..., Any]]] = set()

    def __init__(self, f: Callable[..., Any]) -> None:
        self.func = cache(f)
        update_wrapper(self, f)
        Cache._cached_wrappers.add(self.func)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.func(*args, **kwargs)

    @staticmethod
    @contextmanager
    def manage() -> Generator[None]:
        try:
            yield
        finally:
            Cache.clear()

    @staticmethod
    def clear() -> None:
        for cached_func in Cache._cached_wrappers:
            cached_func.cache_clear()  # type: ignore[attr-defined]


@Cache
def scale_down(value: float, percentage: float, precision: int = 0, min_value: int = 1) -> float:
    """Scales down a value with rounding

    :param value: The current value to calculate the percentage from
    :param percentage: The percentage to apply to the current value
    :param precision: The number of decimal places to round to
    :param min_value: The minimum allowed value after calculation
    """
    if percentage > 100:
        raise ValueError("The percentage must be 100 or smaller")
    val = _round_half_up(value * percentage / 100, precision)
    return max(val, min_value)


@Cache
def generate_group_id(item: Item, scope: str) -> str | None:
    """Generate a smoke scope group ID for the item

    :param item: Collected Pytest item
    :param scope: Smoke scope
    """
    assert scope
    if item.config.hook.pytest_smoke_exclude(item=item, scope=scope):
        return None

    if (group_id := item.config.hook.pytest_smoke_generate_group_id(item=item, scope=scope)) is not None:
        return group_id

    return _generate_scope_group_id(item, scope)


@Cache
def has_parametrized_test(node: Node) -> bool:
    """Check if at least one parametrized test exists in the node

    :param node: Pytest node
    """
    node_items = tuple(x for x in node.session.items if x.parent == node)
    for node_item in node_items:
        if node_item.get_closest_marker("parametrize"):
            return True
    return False


def sort_items(items: list[Item], session: Session, smoke_option: SmokeOption) -> list[Item]:
    """Sort collected Pytest items for the given select mode

    :param items: Collected Pytest items
    :param session: Pytest session
    :param smoke_option: Smoke option
    """
    if smoke_option.select_mode == SmokeSelectMode.FIRST:
        sorted_items = items
    elif smoke_option.select_mode == SmokeSelectMode.LAST:
        sorted_items = items[::-1]
    elif smoke_option.select_mode == SmokeSelectMode.RANDOM:
        random_: random.Random | ModuleType
        if smoke.is_xdist_installed and (is_xdist_controller(session) or is_xdist_worker(session)):
            # Set the seed to ensure XDIST controller and workers collect the same items
            random_ = random.Random(UUID(os.environ[SmokeEnvVar.SMOKE_TEST_SESSION_UUID]).time)
        else:
            random_ = random
        sorted_items = random_.sample(items, len(items))
    else:
        sorted_items = session.config.hook.pytest_smoke_sort_by_select_mode(
            items=items.copy(), scope=smoke_option.scope, select_mode=smoke_option.select_mode
        )
        if sorted_items is None:
            raise pytest.UsageError(
                f"The custom sort logic for the select mode '{smoke_option.select_mode}' must be implemented using the "
                f"pytest_smoke_sort_by_select_mode hook"
            )
    return sorted_items


def parse_n(value: str) -> int | float | str:
    v = value.strip()
    try:
        if is_scale := v.endswith("%"):
            num = float(v[:-1])
        else:
            num = int(v)

        if num < 1 or (is_scale and num > 100):
            raise ValueError

        if is_scale:
            return f"{num}%"
        else:
            return num
    except ValueError:
        raise pytest.UsageError(
            f"The smoke N value must be a positive number or a valid percentage. '{value}' was given."
        )


def parse_select_mode(value: str) -> str:
    if (v := value.strip()) == "":
        raise pytest.UsageError(f"Invalid select mode: '{value}'")
    return v


def parse_scope(value: str) -> str:
    if (v := value.strip()) == "":
        raise pytest.UsageError(f"Invalid scope: '{value}'")
    return v


def parse_ini_option(config: Config, option: SmokeIniOption) -> str | int | float | bool:
    try:
        v = config.getini(option)
        if option == SmokeIniOption.SMOKE_DEFAULT_N:
            return parse_n(v)
        elif option == SmokeIniOption.SMOKE_DEFAULT_SELECT_MODE:
            return parse_select_mode(v)
        elif option == SmokeIniOption.SMOKE_DEFAULT_SCOPE:
            return parse_scope(v)
        else:
            return v
    except ValueError as e:
        raise pytest.UsageError(f"{option}: {e}")


def _round_half_up(x: float, precision: int) -> float:
    return float(Decimal(str(x)).quantize(Decimal("10") ** -precision, rounding=ROUND_HALF_UP))


def _generate_scope_group_id(item: Item, scope: str) -> str | None:
    def generate_class_group_id(current_item: Node, class_id: str = "") -> str:
        parent = current_item.parent
        if isinstance(parent, Class):
            return generate_class_group_id(parent, class_id=f"::{parent.name}{class_id}")
        return class_id

    if scope not in [str(x) for x in SmokeScope]:
        raise pytest.UsageError(
            f"The logic for the custom scope '{scope}' must be implemented using the "
            f"pytest_smoke_generate_group_id hook"
        )

    if scope == SmokeScope.ALL:
        return "*"

    is_class_method = isinstance(item.parent, Class)
    if not is_class_method and scope == SmokeScope.CLASS:
        return None

    file_path = item.path
    if scope == SmokeScope.DIRECTORY:
        return str(file_path.parent)

    group_id = str(file_path)
    if scope == SmokeScope.FILE:
        return group_id

    if is_class_method:
        group_id += generate_class_group_id(item)
        if scope == SmokeScope.CLASS:
            return group_id

    # function or auto scope
    assert scope in [SmokeScope.FUNCTION, SmokeScope.AUTO]
    if scope == SmokeScope.FUNCTION or has_parametrized_test(item.parent):
        func_name = cast(Function, item).function.__name__
        return f"{group_id}::{func_name}"
    else:
        # The parent node has no parametrized tests. Fall back to file or class scope
        if is_class_method:
            dynamic_scope = SmokeScope.CLASS
        else:
            dynamic_scope = SmokeScope.FILE
        return _generate_scope_group_id(item, dynamic_scope)



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any

import pytest
from pytest import Pytester
from pytest_mock import MockerFixture

from pytest_smoke import smoke

if TYPE_CHECKING:
    from pytest import Item


if os.environ.get("IGNORE_XDIST") == "true":
    smoke.is_xdist_installed = False

from pytest_smoke.compat import StrEnum
from tests.helper import TestClassSpec, TestFileSpec, TestFuncSpec, generate_test_code, patch_runpytest

Pytester.runpytest = patch_runpytest(Pytester.runpytest)
pytest_plugins = "pytester"


def pytest_runtest_setup(item: Item) -> None:
    if item.get_closest_marker("xdist") and not smoke.is_xdist_installed:
        pytest.skip(reason="pytest-xdist is required")


@pytest.hookimpl(trylast=True)
def pytest_make_parametrize_id(val: Any, argname: str) -> str:
    if isinstance(val, StrEnum):
        val = str(val)
    return f"{argname}={val!r}"


@pytest.fixture(autouse=True)
def mock_is_xdist_installed(mocker: MockerFixture) -> None:
    """Mock the smoke.is_xdist_installed flag value to limit the effect of a change the plugin will make during some
    tests to be within a test
    """
    mocker.patch.object(smoke, "_is_xdist_installed", return_value=smoke.is_xdist_installed)


@pytest.fixture(scope="session")
def test_file_specs() -> list[TestFileSpec]:
    """Specifications for test files to generate"""
    # file1: Non-parametrized test functions
    test_file_spec1 = TestFileSpec([TestFuncSpec(), TestFuncSpec()])
    # file2: With parametrized test functions
    test_file_spec2 = TestFileSpec([TestFuncSpec(), TestFuncSpec(num_params=10), TestFuncSpec(num_params=20)])
    # file3: Test classes with/without class-level parametrization and with/without nested test classes
    test_file_spec3 = TestFileSpec(
        [
            TestClassSpec("Test1", [TestFuncSpec(), TestFuncSpec()]),
            TestClassSpec("Test2", [TestFuncSpec(), TestFuncSpec(num_params=5), TestFuncSpec(num_params=10)]),
            TestClassSpec("Test3", [TestFuncSpec(), TestFuncSpec(num_params=3)], num_params=2),
            TestClassSpec(
                "Test4",
                [TestFuncSpec(), TestFuncSpec()],
                nested_test_class_specs=[
                    TestClassSpec("TestNested1", [TestFuncSpec(), TestFuncSpec()]),
                    TestClassSpec("TestNested2", [TestFuncSpec(), TestFuncSpec(num_params=3)]),
                    TestClassSpec("TestNested3", [TestFuncSpec(), TestFuncSpec(num_params=2)], num_params=3),
                ],
            ),
            TestClassSpec(
                "Test5",
                [TestFuncSpec(), TestFuncSpec()],
                num_params=2,
                nested_test_class_specs=[
                    TestClassSpec("TestNested1", [TestFuncSpec(), TestFuncSpec()]),
                    TestClassSpec("TestNested2", [TestFuncSpec(), TestFuncSpec(num_params=3)]),
                    TestClassSpec("TestNested3", [TestFuncSpec(), TestFuncSpec(num_params=2)], num_params=3),
                ],
            ),
        ]
    )
    # file4: A mix of everything above, inside a sub directory
    test_file_spec4 = TestFileSpec(
        [*test_file_spec1.test_specs, *test_file_spec2.test_specs, *test_file_spec3.test_specs],
        test_dir="tests_something1",
    )
    return [test_file_spec1, test_file_spec2, test_file_spec3, test_file_spec4]


@pytest.fixture
def generate_test_files(pytester: Pytester, test_file_specs: list[TestFileSpec]) -> None:
    """Generate test files with given test file specs"""
    test_files = {}
    for i, test_file_spec in enumerate(test_file_specs, start=1):
        test_filename = f"test_{i}"
        if test_dir := test_file_spec.test_dir:
            test_filename = test_dir + os.sep + test_filename
        test_files[test_filename] = generate_test_code(test_file_spec)
    pytester.makepyfile(**test_files)



from __future__ import annotations

import sys
from enum import Enum
from typing import Any

import pytest

if sys.version_info < (3, 11):

    class StrEnum(str, Enum):
        def _generate_next_value_(name: str, start: int, count: int, last_values: list[Any]) -> str:  # type: ignore[override]
            return name.lower()

        def __str__(self) -> str:
            return str(self.value)
else:
    from enum import StrEnum  # noqa: F401


if pytest.version_tuple < (7, 4):
    from collections.abc import Mapping
    from typing import NamedTuple

    class TestShortLogReport(NamedTuple):
        category: str
        letter: str
        word: str | tuple[str, Mapping[str, bool]]
else:
    from pytest import TestShortLogReport  # type: ignore[no-redef] # noqa: F401



from __future__ import annotations

import re

import pytest
from pytest import ExitCode, Pytester

from pytest_smoke import smoke
from pytest_smoke.types import SmokeIniOption, SmokeScope, SmokeSelectMode
from tests.helper import (
    TEST_NAME_BASE,
    TestFileSpec,
    TestFuncSpec,
    generate_test_code,
    get_num_tests,
    get_num_tests_to_be_selected,
)


def test_smoke_command_help(pytester: Pytester) -> None:
    """Test pytest command help for this plugin"""
    result = pytester.runpytest("-h")
    assert result.ret == ExitCode.OK, str(result.stderr)
    stdout = str(result.stdout)
    pattern1 = (
        r"\n"
        + r"Smoke testing:\n"
        + (r"  --smoke=\[N\]\s+.+")
        + (
            r"  --smoke-scope=SCOPE\s+.+"
            + r"\n".join(
                rf"\s+- {s + (re.escape(' (default)') if s == SmokeScope.AUTO else '')}: .+" for s in SmokeScope
            )
        )
        + (r"  --smoke-select-mode=MODE\s+.+" + r"\n".join(rf"\s+- {s}: .+" for s in SmokeSelectMode))
        + r"\n"
    )
    assert re.search(pattern1, stdout, re.DOTALL)

    pattern2 = r"\[pytest\] .+" + r"\n".join(rf"  {opt} \([^)]+\):\s+\[pytest-smoke\] .+" for opt in SmokeIniOption)
    assert re.search(pattern2, stdout, re.DOTALL)


def test_smoke_show_markers(pytester: Pytester) -> None:
    """Test the custom marker information provided by the plugin"""
    result = pytester.runpytest("--markers")
    assert result.ret == ExitCode.OK
    result.stdout.re_match_lines([r"@pytest\.mark\.smoke\(\*, mustpass=False, runif=True\): .+"])


@pytest.mark.parametrize("with_xdist", [False, pytest.param(True, marks=pytest.mark.xdist)])
@pytest.mark.usefixtures("generate_test_files")
def test_smoke_no_option(pytester: Pytester, test_file_specs: list[TestFileSpec], with_xdist: bool) -> None:
    """Test the plugin does not affect pytest/pytest-xdist when no plugin options are given"""
    num_all_tests = get_num_tests(*test_file_specs)
    args = []
    if with_xdist:
        args.extend(["-n", "auto"])
    result = pytester.runpytest(*args)
    assert result.ret == ExitCode.OK
    result.assert_outcomes(passed=num_all_tests)


@pytest.mark.usefixtures("generate_test_files")
@pytest.mark.parametrize("scope", [None, *SmokeScope])
@pytest.mark.parametrize("n", [None, "1", "5", "15", "25", "35", str(2**32 - 1), "1%", "1.23%", "10%", "33%", "100%"])
def test_smoke_n(pytester: Pytester, test_file_specs: list[TestFileSpec], n: str | None, scope: str | None) -> None:
    """Test various smoke N values with/without a smoke scope"""
    num_all_tests = get_num_tests(*test_file_specs)
    num_tests_to_be_selected = get_num_tests_to_be_selected(test_file_specs, n, scope)
    args = ["--smoke"]
    if n:
        args.append(n)
    if scope:
        args.extend(["--smoke-scope", scope])
    result = pytester.runpytest(*args)
    assert result.ret == ExitCode.OK
    result.assert_outcomes(passed=num_tests_to_be_selected, deselected=num_all_tests - num_tests_to_be_selected)


@pytest.mark.parametrize("select_mode", [None, *SmokeSelectMode])
def test_smoke_select_mode(pytester: Pytester, select_mode: str | None) -> None:
    """Test the predefined test selection logic with/without the --smoke-select-mode option"""
    smoke_n = 5
    num_tests = 100
    pytester.makepyfile(generate_test_code(TestFuncSpec(num_params=num_tests)))
    args = ["--smoke", str(smoke_n), "--co", "-q"]
    if select_mode:
        args.extend(["--smoke-select-mode", select_mode])

    prev_test_nums = None
    for i in range(2):
        result = pytester.runpytest(*args)
        assert result.ret == ExitCode.OK
        result.assert_outcomes(deselected=num_tests - smoke_n)

        matched_test_nums = re.findall(rf"test_.+\.py::{TEST_NAME_BASE}\[(\d+)\]", str(result.stdout))
        assert len(matched_test_nums) == smoke_n
        test_nums = [int(n) for n in matched_test_nums]

        if select_mode == SmokeSelectMode.LAST:
            assert test_nums == [n for n in range(num_tests)][-smoke_n:]
        elif select_mode == SmokeSelectMode.RANDOM:
            assert sorted(test_nums) == test_nums
        else:
            assert test_nums == [n for n in range(num_tests)][:smoke_n]

        if prev_test_nums:
            if select_mode == SmokeSelectMode.RANDOM:
                assert test_nums != prev_test_nums
            else:
                assert test_nums == prev_test_nums
        else:
            prev_test_nums = test_nums


@pytest.mark.parametrize("num_fails", [0, 1, 2])
@pytest.mark.parametrize("runif", [None, False, True])
@pytest.mark.parametrize("mustpass", [None, False, True])
def test_smoke_marker_critical_tests(pytester: Pytester, mustpass: bool, runif: bool | None, num_fails: int) -> None:
    """Test @pytest.mark.smoke marker with/without the optional mustpass and runif kwargs"""

    def is_critical(i: int) -> bool:
        return bool(i % 2)

    def param_marker(i: int) -> str | None:
        mark = "smoke"
        if is_critical(i):
            kwargs = {}
            if i in test2_pos_mustpass and mustpass is not None:
                kwargs["mustpass"] = mustpass
            if i in test2_pos_with_runif and runif is not None:
                kwargs["runif"] = runif
            if kwargs:
                args = ", ".join(f"{k}={v}" for k, v in kwargs.items())
                mark = f"{mark}({args})"
            return mark
        else:
            return None

    smoke_n = 4
    num_tests_1 = 10
    num_tests_2 = 20
    num_mustpass = 2
    num_critical_with_runif = 3
    test2_pos_critical = list(filter(is_critical, range(num_tests_2)))
    test2_pos_mustpass = test2_pos_critical[:num_mustpass]
    test2_pos_with_runif = test2_pos_critical[-num_critical_with_runif:]
    func_body = f"\tassert {TestFuncSpec.param_arg_name} not in {test2_pos_mustpass[:num_fails]}" if num_fails else None
    test_file_spec = TestFileSpec(
        [
            TestFuncSpec(num_params=num_tests_1),
            TestFuncSpec(num_params=num_tests_2, param_marker=param_marker, func_body=func_body),
        ]
    )
    num_critical_tests = sum([is_critical(i) for i in range(num_tests_2)])
    if runif is False:
        num_critical_tests -= len(test2_pos_with_runif)
    num_regular_tests = smoke_n * len(test_file_spec.test_specs)
    num_expected_selected_tests = num_regular_tests + num_critical_tests
    num_skips = num_regular_tests if mustpass and num_fails else 0
    assert all(
        [
            smoke_n < num_tests_1,
            smoke_n < num_tests_2,
            smoke_n < num_critical_tests,
            smoke_n + num_critical_tests < num_tests_2,
            num_fails <= num_mustpass,
            num_mustpass < num_critical_tests,
            num_critical_with_runif < num_critical_tests,
            all((p < num_tests_2 and is_critical(p)) for p in test2_pos_mustpass),
        ]
    ), "Invalid test conditions"

    pytester.makepyfile(generate_test_code(test_file_spec))
    pytester.makeini(f"""
    [pytest]
    {SmokeIniOption.SMOKE_MARKED_TESTS_AS_CRITICAL} = true
    """)
    result = pytester.runpytest("--smoke", str(smoke_n), "-v")
    assert result.ret == (ExitCode.TESTS_FAILED if num_fails else ExitCode.OK)
    result.assert_outcomes(
        passed=num_expected_selected_tests - (num_fails + num_skips),
        failed=num_fails,
        skipped=num_skips,
        deselected=num_tests_1 + num_tests_2 - num_expected_selected_tests,
    )

    test_name_idx_result_word_pairs = re.findall(
        rf"test_.+\.py::({TEST_NAME_BASE}\d)\[(\d+)\] ((?:PASSED|FAILED|SKIPPED).*?)\s{{2,}}.+", str(result.stdout)
    )
    assert len(test_name_idx_result_word_pairs) == num_expected_selected_tests
    for i, (test_name, param_idx, result_word) in enumerate(test_name_idx_result_word_pairs):
        expected_test_name = (
            f"{TEST_NAME_BASE}1" if num_critical_tests <= i < num_critical_tests + smoke_n else f"{TEST_NAME_BASE}2"
        )
        assert test_name == expected_test_name
        if i < num_critical_tests:
            # ciritical tests
            assert is_critical(int(param_idx))
            word = "FAILED" if num_fails and int(param_idx) in test2_pos_mustpass[:num_fails] else "PASSED"
            if mustpass and int(param_idx) in test2_pos_mustpass:
                assert result_word == word + " (must-pass)"
            else:
                assert result_word == word
        # regular tests
        elif mustpass and num_fails:
            assert (
                result_word == f"SKIPPED ({num_fails}/{len(test2_pos_mustpass)} must-pass "
                f"smoke test{'s' if num_fails > 1 else ''} failed)"
            )
        else:
            assert result_word == "PASSED"


@pytest.mark.xdist
@pytest.mark.usefixtures("generate_test_files")
@pytest.mark.parametrize("select_mode", [None, *SmokeSelectMode])
@pytest.mark.parametrize("scope", [None, *SmokeScope])
def test_smoke_xdist(
    pytester: Pytester, test_file_specs: list[TestFileSpec], scope: str | None, select_mode: str | None
) -> None:
    """Test basic plugin functionality with pytest-xdist.

    The plugin should handle the following points:
    - Report the number of deselected tests in xdist controller
    - When the select mode is random, sets a common random seeds so that all workers can collect the exact same
      tests
    """
    smoke_n = 3
    num_all_tests = get_num_tests(*test_file_specs)
    num_tests_to_be_selected = get_num_tests_to_be_selected(test_file_specs, str(smoke_n), scope)
    args = ["--smoke", str(smoke_n), "-n", "2", "-v"]
    if scope:
        args.extend(["--smoke-scope", scope])
    if select_mode:
        args.extend(["--smoke-select-mode", select_mode])
    result = pytester.runpytest(*args)
    assert result.ret == ExitCode.OK
    result.assert_outcomes(passed=num_tests_to_be_selected, deselected=num_all_tests - num_tests_to_be_selected)


@pytest.mark.xdist
@pytest.mark.parametrize("select_mode", [None, *SmokeSelectMode])
def test_smoke_xdist_disabled(pytester: Pytester, select_mode: str | None) -> None:
    """Test that pytest-smoke does not access the pytest-xdist plugin when it is install but explicitly disabled"""
    assert smoke.is_xdist_installed
    num_tests = 10
    pytester.makepyfile(generate_test_code(TestFuncSpec(num_params=num_tests)))
    args = ["--smoke", "-p", "no:xdist"]
    if select_mode:
        args.extend(["--smoke-select-mode", select_mode])
    result = pytester.runpytest(*args)
    assert result.ret == ExitCode.OK
    assert not smoke.is_xdist_installed
    result.assert_outcomes(passed=1, deselected=num_tests - 1)


@pytest.mark.filterwarnings("ignore::pluggy.PluggyTeardownRaisedWarning")
@pytest.mark.parametrize("n", ["-1", "0", "0.5", "1.1", "foo", " -1%", "0%", "101%", "bar%"])
def test_smoke_invalid_n(pytester: Pytester, n: str) -> None:
    """Test the option with invalid values"""
    result = pytester.runpytest("--smoke", n)
    assert result.ret == ExitCode.USAGE_ERROR
    result.stderr.re_match_lines(
        [rf"ERROR: The smoke N value must be a positive number or a valid percentage\. '{n}' was given"]
    )


@pytest.mark.parametrize("option", ["--smoke-scope", "--smoke-select-mode"])
def test_smoke_without_n_option(pytester: Pytester, option: str) -> None:
    """Test the --smoke option is required to use any functionality provided by the plugin"""
    result = pytester.runpytest(option, "foo")
    assert result.ret == ExitCode.USAGE_ERROR
    result.stderr.re_match_lines([r"ERROR: The --smoke option is requierd to use the pytest-smoke functionality"])



from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("pytest-smoke")
except PackageNotFoundError:
    pass


class PytestSmoke:
    def __init__(self) -> None:
        try:
            from xdist import __version__ as xdist_ver

            self._is_xdist_installed = tuple(map(int, xdist_ver.split("."))) >= (2, 3, 0)
        except ImportError:
            self._is_xdist_installed = False

    @property
    def is_xdist_installed(self) -> bool:
        return self._is_xdist_installed

    @is_xdist_installed.setter
    def is_xdist_installed(self, v: bool) -> None:
        self._is_xdist_installed = v


smoke = PytestSmoke()



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from pytest import hookspec

if TYPE_CHECKING:
    from pytest import Item


@hookspec(firstresult=True)
def pytest_smoke_generate_group_id(item: Item, scope: str) -> Any | None:
    """Return a smoke scope group ID for the predefined or custom scopes

    Use this hook to either override the logic of the predefined scopes or to implement logic for your own scopes
    NOTE: Any custom scopes given to the --smoke-scope option must be handled in this hook
    """


@hookspec(firstresult=True)
def pytest_smoke_include(item: Item, scope: str) -> bool | None:
    """Return True for tests that should be included as "additional" smoke tests

    NOTE: These items will not be counted towards the calculation of N
    """


@hookspec(firstresult=True)
def pytest_smoke_exclude(item: Item, scope: str) -> bool | None:
    """Return True for tests that should not be selected

    Note:
        - These items will not be included in the total number of tests to which N% is applied
        - This hook takes precedence over any other options provided by the plugin

    """


@hookspec(firstresult=True)
def pytest_smoke_sort_by_select_mode(items: list[Item], scope: str, select_mode: str) -> list[Item] | None:
    """Return sorted items to implement a test selection logic for the custom select mode.
    The plugin will pick N tests from each scope group based on the sorted items, meaning that an item appearing
    earlier in the same scope group has a higher chance of being selected.

    Note:
        - The hook is called only for custom select modes
        - This hook does not affect the test execution order

    """
